# Copyright 2016-2018 Martina Stella, Fawzi Mohamed
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from builtins import range
from builtins import object
import numpy as np
import nomadcore.ActivateLogging
from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import mainFunction
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from .OnetepCommon import get_metaInfo
import logging, os, re, sys



############################################################
# This is the parser for the *.md file of Onetep.
############################################################

logger = logging.getLogger("nomad.OnetepTSParser")

class OnetepTSParserContext(object):
    """Context for parsing Onetep *.md file.


    The onClose_ functions allow processing and writing of cached values after a section is closed.
    They take the following arguments:
        backend: Class that takes care of wrting and caching of metadata.
        gIndex: Index of the section that is closed.
        section: The cached values and sections that were found in the section that is closed.
    """
    def __init__(self, writeMetaData = True):
        """Args:
            writeMetaData: Deteremines if metadata is written or stored in class attributes.
        """
        self.writeMetaData = writeMetaData

        self.total_energy = []
        self.total_energy_final = []
        self.total_energy_pro = []

        self.md_forces =[]
        self.md_veloc =[]
        self.frame_atom_label =[]

        self.total_forces =[]
        self.atom_label = []
        self.frame_stress_tensor =[]
        self.total_positions =[]
        self.frame_cell=[]
        self.frame_cell_final=[]
        self.vector_velocities=[]
        self.frame_time =[]
        self.path_ts = []
        self.path_final = []
        self.path_pro = []
        self.total_positions_final=[]
        self.total_positions_pro=[]
    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts and the compiled parser is obtained.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser

    def onClose_x_onetep_section_ts_store(self, backend, gIndex, section):


        vet = section ['x_onetep_ts_cell_vectors_store']
        forces_ts = section ['x_onetep_ts_forces_store']

        position = section ['x_onetep_ts_positions_store']
        energy = section['x_onetep_ts_energy']
        # path_step = section ['x_Onetep_ts_path']

        # for i in range (len(path_step)):
        #     self.path_ts.append(path_step[i])
        bohr_to_m = float(5.29177211e-11)
        Hr_J_converter = float(4.35974e-18)
        HrK_to_K_coverter= float(3.1668114e-6)
        hrang_to_N = float(4.359745e-8)
        for i in energy:

            energy = [x * Hr_J_converter for x in energy]
            self.total_energy.extend(energy)


        if vet:
            self.cell =[]
            for i in range(len(vet)):
                vet[i] = vet[i].split()
                vet[i] = [float(j) for j in vet[i]]
                vet_list = vet[i]
                vet_list = [x * bohr_to_m for x in vet_list]
                self.cell.append(vet_list)
            self.frame_cell.append(self.cell)


        if position:
            self.at_nr = len(position)
            self.atom_position=[]
            for i in range(0, self.at_nr):
                position[i] = position[i].split()
                position[i] = [float(j) for j in position[i]]
                pos_list = position[i]
                pos_list = [x * bohr_to_m  for x in pos_list]
                self.atom_position.append(pos_list)
            self.total_positions.append(self.atom_position)

        if forces_ts is not None:

            self.ts_forces = []
            for f in forces_ts:
                f = f.split()
                f = [float(k) for k in f]
                f_st_intts = f
                f_st_intts = [x * hrang_to_N for x in f_st_intts]
                self.ts_forces.append(f_st_intts)
            self.total_forces.append(self.ts_forces)

    def onClose_x_onetep_section_ts_final_store(self, backend, gIndex, section):
        # path_final_ts = section ['x_Onetep_ts_path_ts_final']
        vet_final = section ['x_onetep_ts_cell_vectors_final_store']
        forces_final = section ['x_onetep_ts_forces_final_store']

        position_final = section ['x_onetep_ts_positions_final_store']
        energy_final = section['x_onetep_ts_energy_final_store']

        # for i in range (len(path_final_ts)):
        #     self.path_final = path_final_ts[i]
        bohr_to_m = float(5.29177211e-11)
        Hr_J_converter = float(4.35974e-18)
        HrK_to_K_coverter= float(3.1668114e-6)

        hrang_to_N = float(4.359745e-8)

        self.total_energy_final = Hr_J_converter * energy_final[0]
        # self.total_energy_final = energy_final


        if vet_final:
            self.cell_final =[]
            for i in range(len(vet_final)):
                vet_final[i] = vet_final[i].split()
                vet_final[i] = [float(j) for j in vet_final[i]]
                vetf_list = vet_final[i]
                vetf_list = [x * bohr_to_m for x in vetf_list]
                self.cell_final.append(vetf_list)
        # self.frame_cell_final.append(self.cell_final)


        if position_final:
            self.at_nr = len(position_final)
            self.atomf_position=[]
            for i in range(0, self.at_nr):
                position_final[i] = position_final[i].split()
                position_final[i] = [float(j) for j in position_final[i]]
                posf_list = position_final[i]
                posf_list = [x * bohr_to_m for x in posf_list]
                self.atomf_position.append(posf_list)
            # self.total_positions_final.append(self.atomf_position)

        if forces_final is not None:

            self.md_forces_final = []
            for f in forces_final:
                f = f.split()
                f = [float(k) for k in f]
                f_st_intf = f
                f_st_intf = [x * hrang_to_N for x in f_st_intf]
                self.md_forces_final.append(f_st_intf)

            # self.total_forces_final.append(self.md_forces_final)

    def onClose_x_onetep_section_ts_product_store(self, backend, gIndex, section):
         # path_product = section ['x_Onetep_ts_path_product']
        vet_pro = section ['x_onetep_ts_cell_vectors_pro_store']
        forces_pro = section ['x_onetep_ts_forces_pro_store']

        position_pro = section ['x_onetep_ts_positions_pro_store']
        energy_pro = section['x_onetep_ts_energy_product_store']

        bohr_to_m = float(5.29177211e-11)
        Hr_J_converter = float(4.35974e-18)
        HrK_to_K_coverter= float(3.1668114e-6)
        hrang_to_N = float(4.359745e-8)
        # for i in range (len(path_product)):
        #     self.path_pro = path_product[i]



        self.total_energy_pro = energy_pro[0] * Hr_J_converter


        if vet_pro:
            self.cell_pro =[]
            for i in range(len(vet_pro)):
                vet_pro[i] = vet_pro[i].split()
                vet_pro[i] = [float(j) for j in vet_pro[i]]
                vetp_list = vet_pro[i]
                vetp_list = [x * bohr_to_m for x in vetp_list]
                self.cell_pro.append(vetp_list)
        # self.frame_cell_final.append(self.cell_final)


        if position_pro:
            self.at_nr = len(position_pro)
            self.atomp_position=[]
            for i in range(0, self.at_nr):
                position_pro[i] = position_pro[i].split()
                position_pro[i] = [float(j) for j in position_pro[i]]
                posp_list = position_pro[i]
                posp_list = [x * bohr_to_m for x in posp_list]
                self.atomp_position.append(posp_list)
            # self.total_positions_final.append(self.atomf_position)

        if forces_pro is not None:

            self.md_forces_pro = []
            for f in forces_pro:
                f = f.split()
                f = [float(k) for k in f]
                f_st_intp = f
                f_st_intp = [x * hrang_to_N for x in f_st_intp]
                self.md_forces_pro.append(f_st_intp)

    def onClose_section_run(self, backend, gIndex, section):
        path_product = section ['x_onetep_ts_path_product']
        path_final_ts = section ['x_onetep_ts_path_ts_final']
        path_step = section ['x_onetep_ts_path']
        if path_step:
            for i in range (len(path_step)):
                self.path_ts.append(path_step[i])
        if path_product:
            for i in range (len(path_product)):
                self.path_pro = path_product[i]
        if path_final_ts:
            for i in range (len(path_final_ts)):
                self.path_final = path_final_ts[i]

def build_OnetepTSFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the *.md file of Onetep.

    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       SimpleMatcher that parses *.md file of Onetep.
    """
    return SM (name = 'Root1',
        startReStr = "",
        sections = ['section_run'],
        forwardMatch = True,
        weak = True,
        subMatchers = [
            SM (name = 'Root2',
            startReStr =r"\sLST\s*[0-9.]\s*(?P<x_onetep_ts_path>[-+0-9.eEdD]+)\s*",
            endReStr ="/n",
            sections = ['x_onetep_section_ts_store'],
            repeats = True,
            subMatchers = [
                SM (r"\s*(?P<x_onetep_ts_energy>[-+0-9.eEdD]+)\s*[-+0-9.eEdD]+\s*\<\-\-\sE\s*"),
                SM (r"\s*(?P<x_onetep_ts_cell_vectors_store>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sh\s",repeats = True),
                SM(r"\s*[A-Za-z]+\s*[0-9.]+\s*(?P<x_onetep_ts_positions_store>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sR\s*",repeats = True),
                SM(r"\s*[A-Za-z]+\s*[0-9.]+\s*(?P<x_onetep_ts_forces_store>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sF\s*",repeats = True,endReStr ="/n"),

                   ]),
            SM (name = 'Root3',
            startReStr =r"\sQST\s*[0-9.]\s*(?P<x_onetep_ts_path>[-+0-9.eEdD]+)\s*",
            endReStr ="/n",
            sections = ['x_onetep_section_ts_store'],
            repeats = True,
            subMatchers = [
                SM (r"\s*(?P<x_onetep_ts_energy>[-+0-9.eEdD]+)\s*[-+0-9.eEdD]+\s*\<\-\-\sE\s*"),
                SM (r"\s*(?P<x_onetep_ts_cell_vectors_store>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sh\s",repeats = True),
                SM(r"\s*[A-Za-z]+\s*[0-9.]+\s*(?P<x_onetep_ts_positions_store>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sR\s*",repeats = True),
                SM(r"\s*[A-Za-z]+\s*[0-9.]+\s*(?P<x_onetep_ts_forces_store>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sF\s*",repeats = True,endReStr ="/n"),

                   ]),
            SM (name = 'Root4',
                startReStr =r"\sTS\s*0\s*(?P<x_onetep_ts_path_ts_final>[-+0-9.eEdD]+)\s*",
                endReStr ="/n",
                sections = ['x_onetep_section_ts_final_store'],
                repeats = True,
                subMatchers = [
                            SM (r"\s*(?P<x_onetep_ts_energy_final_store>[-+0-9.eEdD]+)\s*[-+0-9.eEdD]+\s*\<\-\-\sE\s*"),
                            SM (r"\s*(?P<x_onetep_ts_cell_vectors_final_store>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sh\s",repeats = True),
                            SM(r"\s*[A-Za-z]+\s*[0-9.]+\s*(?P<x_onetep_ts_positions_final_store>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sR\s*",repeats = True),
                            SM(r"\s*[A-Za-z]+\s*[0-9.]+\s*(?P<x_onetep_ts_forces_final_store>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sF\s*",repeats = True,endReStr ="/n"),
                    ]),
            SM (name = 'Root5',
                startReStr =r"\sPRO\s*0\s*(?P<x_onetep_ts_path_product>[-+0-9.eEdD]+)\s*",
                endReStr ="/n",
                sections = ['x_onetep_section_ts_product_store'],
                repeats = True,
                subMatchers = [
                            SM (r"\s*(?P<x_onetep_ts_energy_product_store>[-+0-9.eEdD]+)\s*[-+0-9.eEdD]+\s*\<\-\-\sE\s*"),
                            SM (r"\s*(?P<x_onetep_ts_cell_vectors_pro_store>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sh\s",repeats = True),
                            SM(r"\s*[A-Za-z]+\s*[0-9.]+\s*(?P<x_onetep_ts_positions_pro_store>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sR\s*",repeats = True),
                            SM(r"\s*[A-Za-z]+\s*[0-9.]+\s*(?P<x_onetep_ts_forces_pro_store>[-+0-9.eEdD]+\s*[-+0-9.eEdD]+\s*[-+0-9.eEdD]+)\s*\<\-\-\sF\s*",repeats = True,endReStr ="/n"),
                    ]),

            ])





def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
                                 'section_run': CachingLvl,
                                 # 'x_onetep_section_ts_store': CachingLvl,
                                 # 'x_onetep_section_ts_final_store': CachingLvl,
                                 # 'x_onetep_section_ts_product_store': CachingLvl,
                              }
    # Set all band metadata to Cache as they need post-processsing.
    for name in metaInfoEnv.infoKinds:
        if name.startswith('x_onetep_'):
            cachingLevelForMetaName[name] = CachingLevel.Cache
    return cachingLevelForMetaName


def main(CachingLvl):
    """Main function.

    Set up everything for the parsing of the Onetep *.cell file and run the parsing.

    Args:
        CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
            This allows to run the parser without opening new sections.
    """
    # get band.out file description
    OnetepTSFileSimpleMatcher = build_OnetepTSFileSimpleMatcher()
    # loading metadata from nomad-meta-info/meta_info/nomad_meta_info/Onetep.nomadmetainfo.json
    metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../../../../nomad-meta-info/meta_info/nomad_meta_info/onetep.nomadmetainfo.json"))
    metaInfoEnv = get_metaInfo(metaInfoPath)
    # set parser info
    parserInfo = {'name':'Onetep-ts-parser', 'version': '1.0'}
    # get caching level for metadata
    cachingLevelForMetaName = get_cachingLevelForMetaName(metaInfoEnv, CachingLvl)
    # start parsing
    mainFunction(mainFileDescription = OnetepTSFileSimpleMatcher,
                 metaInfoEnv = metaInfoEnv,
                 parserInfo = parserInfo,
                 cachingLevelForMetaName = cachingLevelForMetaName,
                 superContext = OnetepTSParserContext())

if __name__ == "__main__":
    main(CachingLevel.Forward)



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' The beginning of a pylint plugin. Unfotunately it is kinda nonsensical without a
partnering mypy plugin. '''

import astroid
from astroid import MANAGER


annotation_names = {
    'MSection': {
        'a_test': '*'
    },
    'Section': {
        'a_elastic': 'nomad.metainfo.elastic_extension.ElasticDocument',
        'a_mongo': 'nomad.metainfo.mongoengine_extension.MongoDocument'
    },
    'Quantity': {
        'a_elastic': 'nomad.metainfo.elastic_extension.Elastic'
    }
}


def register(linter):
    # Needed for registering the plugin.
    pass


def transform(cls):
    ''' Transforms annotation fields for known annotation classes. '''
    for cls_name, annotations in annotation_names.items():
        if cls.name == cls_name:
            for name, type_spec in annotations.items():
                if type_spec == '*':
                    cls.locals[name] = [astroid.Instance()]
                else:
                    type_path = type_spec.split('.')
                    type_module = '.'.join(type_path[:-1])
                    type_name = type_path[-1]
                    module = MANAGER.ast_from_module_name(type_module)
                    cls.locals[name] = [cls.instantiate_class() for cls in module.lookup(type_name)[1]]


def is_derived(func):
    ''' Check if this is call to the derived decorator. '''
    decorators = func.decorators.nodes if func.decorators else []
    for decorator_node in decorators:
        if isinstance(decorator_node, astroid.Call):
            if decorator_node.func and isinstance(decorator_node.func, astroid.Name):
                return decorator_node.func.name == 'derived'

    return False


def derived_transform(node, context=None):
    '''
    The derived decorator produces a Quantity. Pylint does not infer this on its own.
    We change the inferred type of a @derived call to a Quantity instance here.
    '''
    module = MANAGER.ast_from_module_name('nomad.metainfo.metainfo')
    class_defs = [cls.instantiate_class() for cls in module.lookup('Quantity')[1]]
    return iter([class_defs[0].instantiate_class()])


MANAGER.register_transform(astroid.ClassDef, transform)
MANAGER.register_transform(astroid.FunctionDef, astroid.inference_tip(derived_transform), is_derived)



# Copyright 2016-2018 Fawzi Mohamed, Honghui Shang, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from builtins import object
import numpy as np
import logging
import sys

from nomadcore.caching_backend import CachingLevel
from nomadcore.simple_parser import mainFunction
from nomadcore.simple_parser import SimpleMatcher as SM


############################################################
# This is the parser for the main file of dmol3.
############################################################


############################################################
###############[1] transfer PARSER CONTEXT #################
############################################################
logger = logging.getLogger("nomad.dmol3Parser")

class Dmol3ParserContext(object):

    def __init__(self):
        self.functionals                       = []

    def initialize_values(self):
        """Initializes the values of certain variables.

        This allows a consistent setting and resetting of the variables,
        when the parsing starts and when a section_run closes.
       """
        self.secMethodIndex = None
        self.secSystemDescriptionIndex = None

        self.singleConfCalcs = []


    def startedParsing(self, fInName, parser):
        """Function is called when the parsing starts.

        Get compiled parser, filename and metadata.

        Args:
            fInName: The file name on which the current parser is running.
            parser: The compiled parser. Is an object of the class SimpleParser in nomadcore.simple_parser.py.
        """
        self.parser = parser
        self.fName = fInName
        # save metadata
        self.metaInfoEnv = self.parser.parserBuilder.metaInfoEnv
        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()

    def onClose_section_run(self, backend, gIndex, section):
        """Trigger called when section_run is closed.
        """
        # reset all variables
        self.initialize_values()
        # frame sequence
        sampling_method = "geometry_optimization"

        samplingGIndex = backend.openSection("section_sampling_method")
        backend.addValue("sampling_method", sampling_method)
        backend.closeSection("section_sampling_method", samplingGIndex)
        frameSequenceGIndex = backend.openSection("section_frame_sequence")
        backend.addValue("frame_sequence_to_sampling_ref", samplingGIndex)
        backend.addArrayValues("frame_sequence_local_frames_ref", np.asarray(self.singleConfCalcs))
        backend.closeSection("section_frame_sequence", frameSequenceGIndex)



    def onOpen_section_method(self, backend, gIndex, section):
        # keep track of the latest method section
        self.secMethodIndex = gIndex


    def onOpen_section_system(self, backend, gIndex, section):
        # keep track of the latest system description section
        self.secSystemDescriptionIndex = gIndex



    ###################################################################
    # (2.1) onClose for INPUT geometry (section_system)
    ###################################################################
    def onClose_section_system(self, backend, gIndex, section):
        """Trigger called when section_system is closed.
        Writes atomic positions, atom labels and lattice vectors.
        """
        # keep track of the latest system description section
        self.secSystemDescriptionIndex = gIndex

        atom_pos = []
        for i in ['x', 'y', 'z']:
            api = section['dmol3_geometry_atom_positions_' + i]
            if api is not None:
               atom_pos.append(api)
        if atom_pos:
            # need to transpose array since its shape is [number_of_atoms,3] in the metadata
           backend.addArrayValues('atom_positions', np.transpose(np.asarray(atom_pos)))
            # write atom labels
        atom_labels = section['dmol3_geometry_atom_labels']
        if atom_labels is not None:
           backend.addArrayValues('atom_labels', np.asarray(atom_labels))

        backend.addArrayValues("configuration_periodic_dimensions", np.ones(3, dtype=bool))

        #atom_hirshfeld_population_analysis = section['dmol3_hirshfeld_population']
        #if atom_hirshfeld_population_analysis is not None:
        #   backend.addArrayValues('atom_hirshfeld_population',np.asarray(atom_hirshfeld_population_analysis))
        ###---???shanghui want to know how to add

    def onOpen_section_single_configuration_calculation(self, backend, gIndex, section):
        self.singleConfCalcs.append(gIndex)

    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
# write the references to section_method and section_system
        backend.addValue('single_configuration_to_calculation_method_ref', self.secMethodIndex)
        backend.addValue('single_configuration_calculation_to_system_ref', self.secSystemDescriptionIndex)




    #################################################################
    # (2.2) onClose for INPUT control (section_method)
    #################################################################
    def onClose_section_method(self, backend, gIndex, section):
        functional = section["dmol3_functional_name"]
        if functional:
            functionalMap = {
                "gga": ["GGA_X_PW91","GGA_C_PW91"]
            }
            # Push the functional string into the backend
            nomadNames = functionalMap.get(functional[0])
            if not nomadNames:
                raise Exception("Unhandled xc functional %s found" % functional)
            for name in nomadNames:
                s = backend.openSection("section_XC_functionals")
                backend.addValue('XC_functional_name', name)
                backend.closeSection("section_XC_functionals", s)


    # #################################################################
    # # (3.1) onClose for OUTPUT SCF (section_scf_iteration)
    # #################################################################
    # # Storing the total energy of each SCF iteration in an array
    # def onClose_section_scf_iteration(self, backend, gIndex, section):
    #     """trigger called when _section_scf_iteration is closed"""
    #     # get cached values for energy_total_scf_iteration
    #     ev = section['energy_total_scf_iteration']
    #     self.scfIterNr = len(ev)
    #
    #     #self.energy_total_scf_iteration_list.append(ev)
    #     #backend.addArrayValues('energy_total_scf_iteration_list', np.asarray(ev))
    #     #backend.addValue('number_of_scf_iterations', self.scfIterNr)
    #     #-----???shanghui want to know why can not add them.


    #################################################################
    # (3.2) onClose for OUTPUT eigenvalues (section_eigenvalues)
    #################################################################
    def onClose_section_eigenvalues(self, backend, gIndex, section):
        """Trigger called when _section_eigenvalues is closed.
        Eigenvalues are extracted.
        """
        occs = []
        evs =  []

        ev = section['dmol3_eigenvalue_eigenvalue']
        if ev is not None:
           occ = section['dmol3_eigenvalue_occupation']
           occs.append(occ)
           evs.append(ev)

        self.eigenvalues_occupation = []
        self.eigenvalues_values = []

        #self.eigenvalues_kpoints = []
        self.eigenvalues_occupation.append(occs)
        self.eigenvalues_values.append(evs)




#############################################################
#################[2] MAIN PARSER STARTS HERE  ###############
#############################################################

def build_Dmol3MainFileSimpleMatcher():
    """Builds the SimpleMatcher to parse the main file of dmol3.

    First, several subMatchers are defined, which are then used to piece together
    the final SimpleMatcher.
    SimpleMatchers are called with 'SM (' as this string has length 4,
    which allows nice formating of nested SimpleMatchers in python.

    Returns:
       SimpleMatcher that parses main file of dmol3.
    """


    #####################################################################
    # (1) submatcher for header
    #####################################################################
    headerSubMatcher = SM(name = 'ProgramHeader',
                  startReStr = r"\s*Materials Studio DMol\^3 version (?P<program_version>[0-9.]+)",
                  subMatchers = [
                     SM(r"\s*compiled on\s+(?P<dmol3_program_compilation_date>[a-zA-Z]+\s+[0-9]+\s+[0-9]+)\s+(?P<dmol3_program_compilation_time>[0-9:]*)")
                                  ])

    #####################################################################
    # (2.1) submatcher for INPUT geometry(section_system)
    #####################################################################
    geometrySubMatcher = SM(name = 'Geometry',
        startReStr = r"\s*INCOOR, atomic coordinates in au \(for archive\):",
        sections = ['section_system'],
        subMatchers = [
       # SM (startReStr = r"\s*\|\s*Unit cell:",
       #     subMatchers = [
       #     SM (r"\s*\|\s*(?P<dmol3_geometry_lattice_vector_x__bohr>[-+0-9.]+)\s+(?P<dmol3_geometry_lattice_vector_y__bohr>[-+0-9.]+)\s+(?P<dmol3_geometry_lattice_vector_z__bohr>[-+0-9.]+)", repeats = True)
       #     ]),
        SM (startReStr = r"\s*\$coordinates",
            subMatchers = [
            SM (r"\s*(?P<dmol3_geometry_atom_labels>[a-zA-Z]+)\s+(?P<dmol3_geometry_atom_positions_x__bohr>[-+0-9.]+)\s+(?P<dmol3_geometry_atom_positions_y__bohr>[-+0-9.]+)\s+(?P<dmol3_geometry_atom_positions_z__bohr>[-+0-9.]+)", repeats = True)
            ])
        ])




    ####################################################################
    # (2.2) submatcher for INPUT control (section_method)
    ####################################################################
    calculationMethodSubMatcher = SM(name = 'calculationMethods',
        startReStr = r"\s*INPUT_DMOL keywords \(for archive\):",
        endReStr = r"\s*\>8",
        sections = ["section_method"],
        subMatchers = [
            SM(r"\s*Calculate\s+(?P<dmol3_calculation_type>[A-Za-z_]+)"),
            SM(r"\s*Functional\s+(?P<dmol3_functional_name>[A-Za-z0-9]+)"),
            SM(r"\s*Pseudopotential\s+(?P<dmol3_pseudopotential_name>[A-Za-z]+)"),
            SM(r"\s*Basis\s+(?P<dmol3_basis_name>[A-Za-z]+)"),
            SM(r"\s*Spin_Polarization\s+(?P<dmol3_spin_polarization>[A-Za-z]+)"),
            SM(r"\s*Spin\s+(?P<dmol3_spin>[0-9]+)"),
            SM(r"\s*Atom_Rcut\s+(?P<dmol3_rcut>[-+0-9.eEdD]+)"),
            SM(r"\s*Integration_Grid\s+(?P<dmol3_integration_grid>[A-Za-z]+)"),
            SM(r"\s*Aux_Partition\s+(?P<dmol3_aux_partition>[0-9]+)"),
            SM(r"\s*Aux_Density\s+(?P<dmol3_aux_density>[A-Za-z]+)"),

            SM(r"\s*Charge\s+(?P<dmol3_charge>[-+0-9.eEdD]+)"),
            SM(r"\s*Symmetry\s+(?P<dmol3_symmetry>[A-Za-z0-9]+)"),
            SM(r"\s*Mulliken_Analysis\s+(?P<dmol3_mulliken_analysis>[A-Za-z]+)"),
            SM(r"\s*Hirshfeld_Analysis\s+(?P<dmol3_hirshfeld_analysis>[A-Za-z]+)"),
            SM(r"\s*Partial_Dos\s+(?P<dmol3_partial_dos>[A-Za-z]+)"),
            SM(r"\s*Electrostatic_Moments\s+(?P<dmol3_electrostatic_moments>[A-Za-z]+)"),
            SM(r"\s*Nuclear_EFG\s+(?P<dmol3_nuclear_efg>[A-Za-z]+)"),
            SM(r"\s*Optical_Absorption\s+(?P<dmol3_optical_absorption>[A-Za-z]+)"),
            SM(r"\s*Kpoints\s+(?P<dmol3_kpoints>[A-Za-z]+)"),

            SM(r"\s*SCF_Density_Convergence\s+(?P<dmol3_scf_density_convergence>[-+0-9.eEdD]+)"),
            SM(r"\s*SCF_Spin_Mixing\s+(?P<dmol3_scf_spin_mixing>[-+0-9.eEdD]+)"),
            SM(r"\s*SCF_Charge_Mixing\s+(?P<dmol3_scf_charge_mixing>[-+0-9.eEdD]+)"),
            SM(r"\s*SCF_DIIS\s+(?P<dmol3_scf_diis_number>[-+0-9.eEdD]+)\s+(?P<dmol3_scf_diis_name>[A-Za-z]+)"),
            SM(r"\s*SCF_Iterations\s+(?P<dmol3_scf_iterations>[0-9]+)"),
            SM(r"\s*SCF_Number_Bad_Steps\s+(?P<dmol3_scf_number_bad_steps>[0-9]+)"),
            SM(r"\s*SCF_Direct\s+(?P<dmol3_scf_direct>[A-Za-z]+)"),
            SM(r"\s*SCF_Restart\s+(?P<dmol3_scf_restart>[A-Za-z]+)"),
            SM(r"\s*Occupation\s+(?P<dmol3_occupation_name>[A-Za-z_]+)\s+(?P<dmol3_occupation_width>[0-9.eEdD]+)"),

            SM(r"\s*OPT_Energy_Convergence\s+(?P<dmol3_opt_energy_convergence>[-+0-9.eEdD]+)"),
            SM(r"\s*OPT_Gradient_Convergence\s+(?P<dmol3_opt_gradient_convergence>[-+0-9.eEdD]+)"),
            SM(r"\s*OPT_Displacement_Convergence\s+(?P<dmol3_opt_displacement_convergence>[-+0-9.eEdD]+)"),
            SM(r"\s*OPT_Iterations\s+(?P<dmol3_opt_iterations>[0-9]+)"),
            SM(r"\s*OPT_Coordinate_System\s+(?P<dmol3_opt_coordinate_system>[A-Za-z_]+)"),
            SM(r"\s*OPT_Gdiis\s+(?P<dmol3_opt_gdiis>[A-Za-z]+)"),
            SM(r"\s*OPT_Max_Displacement\s+(?P<dmol3_opt_max_displacement>[-+0-9.eEdD]+)"),
            SM(r"\s*OPT_Steep_Tol\s+(?P<dmol3_opt_steep_tol>[-+0-9.eEdD]+)"),
            SM(r"\s*OPT_Hessian_Project\s+(?P<dmol3_opt_hessian_project>[A-Za-z]+)")


        ])

    ####################################################################
    # (3.1) submatcher for OUPUT SCF
    ####################################################################
    scfSubMatcher = SM(name = 'ScfIterations',
        startReStr = r"\s*Message: Start SCF iterations\s*",
        endReStr = r"\s*Message: SCF converged\s*",
        subMatchers = [
            SM(r"\s*Ef\s+(?P<energy_total_scf_iteration__hartree>[-+0-9.eEdD]+)\s+(?P<dmol3_binding_energy_scf_iteration__hartree>[-+0-9.eEdD]+)\s+(?P<dmol3_convergence_scf_iteration>[-+0-9.eEdD]+)\s+(?P<dmol3_time_scf_iteration>[0-9.eEdD]+)\s+(?P<dmol3_number_scf_iteration>[0-9]+)\s*",
               sections = ['section_scf_iteration'],
               repeats = True)

        ])

    ####################################################################
    # (3.2) submatcher for OUPUT eigenvalues
    ####################################################################
    eigenvalueSubMatcher = SM(name = 'Eigenvalues',
        startReStr = r"\s*state\s+eigenvalue\s+occupation\s*",
        sections = ['section_eigenvalues'],
        subMatchers = [
            SM(r"\s*[0-9]+\s+[+-]\s+[0-9]+\s+[A-Za-z]+\s+[-+0-9.eEdD]+\s+(?P<dmol3_eigenvalue_eigenvalue__eV>[-+0-9.eEdD]+)\s+(?P<dmol3_eigenvalue_occupation>[0-9.eEdD]+)", repeats = True)
        ])


    ####################################################################
    # (3.3) submatcher for OUPUT totalenergy
    ####################################################################
    totalenergySubMatcher = SM(name = 'Totalenergy',
        startReStr = r"\s*Optimization Cycle",
        subMatchers = [
            SM(r"\s*opt==\s+[0-9]+\s+(?P<energy_total__hartree>[-+0-9.eEdD]+)\s+[-+0-9.eEdD]+\s+[-+0-9.eEdD]+\s+[-+0-9.eEdD]+", repeats = True)
        ])


    #####################################################################
    # (3.4) submatcher for OUTPUT relaxation_geometry(section_system)
    #####################################################################
    geometryrelaxationSubMatcher = SM(name = 'GeometryRelaxation',
        startReStr = r"\s*df\s*ATOMIC\s*COORDINATES\s*\(au\)\s*DERIVATIVES\s*\(au\)",
        #endReStr = r"\s*\+\+\+\s+Entering Vibrations Section\s+\+\+\+ ",
        sections = ['section_system'],
        subMatchers = [
        SM (startReStr = r"\s*df\s+x\s+y\s+z\s+x\s+y\s+z",
            subMatchers = [
            SM (r"\s*df\s+(?P<dmol3_geometry_atom_labels>[a-zA-Z]+)\s+(?P<dmol3_geometry_atom_positions_x__angstrom>[-+0-9.]+)\s+(?P<dmol3_geometry_atom_positions_y__angstrom>[-+0-9.]+)\s+(?P<dmol3_geometry_atom_positions_z__angstrom>[-+0-9.]+)\s+[-+0-9.]+\s+[-+0-9.]+\s+[-+0-9.]+", repeats = True)
            ])
        ])


    #####################################################################
    # (3.5) submatcher for OUTPUT population analysis (section_system)
    #####################################################################
    populationSubMatcher = SM(name = 'PopulationAnalysis',
        startReStr = r"\s*\+\+\+\s+Entering Properties Section\s+\+\+\+",
        subMatchers = [
        SM (startReStr = r"\s*Charge partitioning by Hirshfeld method:",
            sections = [ "dmol3_section_hirshfeld_population"],
            subMatchers = [
            SM (r"\s*[a-zA-Z]+\s+[0-9]+\s+charge\s+(?P<dmol3_hirshfeld_population>[-+0-9.]+)", repeats = True)
            ]),
        SM (startReStr = r"\s*Mulliken atomic charges:",
            sections = [ "dmol3_section_mulliken_population"],
            subMatchers = [
            SM (r"\s*[a-zA-Z(]+\s+[0-9)]+\s+(?P<dmol3_mulliken_population>[-+0-9.]+)", repeats = True)
            ])
        ])


    ########################################
    # return main Parser
    ########################################
    return SM (name = 'Root',

        startReStr = "",
        forwardMatch = True,
        weak = True,
        subMatchers = [
        SM (name = 'NewRun',
            startReStr = r"\s*Materials Studio DMol\^3 version [0-9.]",
            endReStr = r"\s*DMol3 job finished successfully",
            repeats = True,
            required = True,
            forwardMatch = True,
            fixedStartValues={'program_name': 'DMol3', 'program_basis_set_type': 'numeric AOs'},
            sections = ['section_run'],
            subMatchers = [

             #-----------(1) header---------------------
             headerSubMatcher,


             #----------(2.1) INPUT : geometry----------
             geometrySubMatcher,
             #----------(2.2) INPUT : control-----------
             calculationMethodSubMatcher,

             SM(name = "single configuration matcher",
                startReStr = r"\s*~~~~~~~~*\s*Start Computing SCF Energy/Gradient\s*~~~~~~~~~~*",
                #endReStr = r"\s*~~~~~~~~*\s*End Computing SCF Energy/Gradient\s*~~~~~~~~~~*",
                repeats = True,
                 sections = ['section_single_configuration_calculation'],
                 subMatchers = [
                    #----------(3.1) OUTPUT : SCF----------------------
                    scfSubMatcher,
                    #----------(3.2) OUTPUT : eigenvalues--------------
                    eigenvalueSubMatcher,
                    #----------(3.4) OUTPUT : relaxation_geometry----------------------
                    geometryrelaxationSubMatcher,
                    ###---???shanghui find this will mismacth the frequencies geometry.
                    #----------(3.3) OUTPUT : totalenergy--------------
                    totalenergySubMatcher,
                    #----------(3.5) OUTPUT : population_analysis----------------------
                    populationSubMatcher
                    ]
                ),


                    #----------(3.6) OUTPUT : frequencies----------------------

           ]) # CLOSING SM NewRun

        ]) # END Root

def get_cachingLevelForMetaName(metaInfoEnv):
    """Sets the caching level for the metadata.

    Args:
        metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.

    Returns:
        Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
                                'eigenvalues_values': CachingLevel.Cache,
                                'eigenvalues_kpoints':CachingLevel.Cache
                                }

    # Set caching for temparary storage variables
    for name in metaInfoEnv.infoKinds:
        if (   name.startswith('dmol3_store_')
            or name.startswith('dmol3_cell_')):
            cachingLevelForMetaName[name] = CachingLevel.Cache
    return cachingLevelForMetaName


# get main file description
Dmol3MainFileSimpleMatcher = build_Dmol3MainFileSimpleMatcher()


class Dmol3Parser():
    """ A proper class envolop for running this parser from within python. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
        from unittest.mock import patch
        logging.debug('dmol3 parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("dmol3.nomadmetainfo.json")
        with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
            mainFunction(
                mainFileDescription=Dmol3MainFileSimpleMatcher,
                metaInfoEnv=None,
                parserInfo={'name':'dmol3-parser', 'version': '1.0'},
                cachingLevelForMetaName=get_cachingLevelForMetaName(backend.metaInfoEnv()),
                superContext=Dmol3ParserContext(),
                superBackend=backend)

        return backend




def get_unitDict(keyname):
    """ Unit dictionary for convertions

        Unit names will be converted to values.
        When defining units in translator dictionary,
        the unit names in the dictionary should be used.
        The unit convertion values are written for SI units.
        If you would like to change it, just add another key
        to the dictionary and change the key at parser base.
        Usage:
           Natively support python language in definitions.
           You can use any python math operator and function.
           Moreover, you can use space or - for multiplication
           and ^ for power of values.
        Example:
            kilogram/meter^2 can be written as
            kilo-gram/meter^2 or kilo-gram/meter**2
            and will be calculated as
            kilo*gram/meter**2

    """
    unitDict = {
        "si" : {
            "meter"          : "1.0",
            "kilo"           : "1.0e3",
            "gram"           : "1.0e-3",
            "second"         : "1.0",
            "joule"          : "1.0",
            "newton"         : "1.0",
            "kelvin"         : "1.0",
            "pascal"         : "1.0",
            "coulomb"        : "1.0",
            "volt"           : "1.0",
            "centi"          : "1.0e-2",
            "milli"          : "1.0e-3",
            "micro"          : "1.0e-6",
            "nano"           : "1.0e-9",
            "pico"           : "1.0e-12",
            "femto"          : "1.0e-15",
            "atto"           : "1.0e-18",
            "erg"            : "1.0e-7",
            "dyne"           : "1.0e-5",
            "bar"            : "1.0e-1",
            "angstrom"       : "1.0e-10",
            "kcal"           : "4184.096739614824",
            "mol"            : "0.602213737699784e24",
            "atmosphere"     : "1.01325e5",
            "electron"       : "1.602176565e-19",
            "atomicmassunit" : "1.66054e-27",
            "amu"            : "1.66054e-27",
            "bohr"           : "5.29177249e-11",
            "hartree"        : "4.35974e-18",
            "pascal"         : "1.0",
            "akmatime"       : "0.048888e-12",
            },
        "amber" : {
            "time"           : "pico*second",
            "length"         : "nano*meter",
            "temperature"    : "Kelvin",
            "energy"         : "kcal/mol",
            },
        "gromacs" : {
            "time"           : "pico*second",
            "length"         : "nano*meter",
            "temperature"    : "Kelvin",
            "energy"         : "kcal/mol",
            },
        "mdtraj" : {
            "time"           : "pico*second",
            "length"         : "nano*meter",
            "angle"          : "degree",
            },
        "mdanalysis" : {
            "time"           : "pico*second",
            "length"         : "angstrom",
            "energy"         : "kilo*joule/mol",
            "charge"         : "electron",
            "force"          : "kilo*joule/(mol*angstrom)",
            "velocity"       : "angstrom/(pico*second)",
            },
        "ase" : {
            "time"           : "pico*second",
            "length"         : "angstrom",
            "energy"         : "kilo*joule/mol",
            "charge"         : "electron",
            "force"          : "kilo*joule/(mol*angstrom)",
            "velocity"       : "angstrom/(pico*second)",
            },
        #AKMA units: http://www.esi.umontreal.ca/accelrys/life/insight2000.1/charmm_principles/Ch01_intro.FM5.html
        "akma" : {
            "time"           : "0.048888e-12*second",
            "outputtime"     : "pico*second",
            "length"         : "angstrom",
            "energy"         : "kcal/mol",
            "charge"         : "electron",
            "mass"           : "1.661e-27*kilo-gram",
            "force"          : "kcal/(mol*angstrom)",
            "velocity"       : "angstrom/akmatime",
            },
        }
    if keyname:
        resDict = unitDict[keyname]
    else:
        resDict = unitDict["si"]
    return resDict

def metaNameConverter(keyName):
    newName = keyName.lower().replace(" ", "").replace("-", "")
    newName = newName.replace("(", "").replace(")", "")
    newName = newName.replace("[", "").replace("]", "")
    newName = newName.replace(",", "").replace(".", "")
    newName = newName.replace("\\", "").replace("/", "")
    newName = newName.replace("'", "").replace(":", "")
    return newName

def metaNameConverter_UnderscoreSpaceDash(keyName):
    newName = keyName.lower()
    newName = ' '.join(newName.split())
    newName = newName.replace(" ", "_").replace("-", "_")
    newName = metaNameConverter(newName)
    return newName

class MetaInfoMap(dict):
    """Map cache values to meta info
    """
    activeInfo=False # The key will be activated when it is matched
    infoPurpose=None # Same as in FileMapDict
    defaultValue=None # Default value of this key
    changeTags=None # Also set the keys in the dict to given values with the match
    nameTranslate=None # translate the key to meta name using this function
    matchStr=None # this string will be matched (can be altered after initialized)
    matchNames=None # Control list/tuple/dict/str for matchStr (See getDict_MetaStrInDict)
    nextMatch=None # Do not alter the match of item until the nextMatch string is matched
    concatMatch=None # If there is an other match in list (Ex.:nextMatch) concatanate it
    replaceTag=None # Replace the original tag with this value but still use the original matchStr
    matchWith=None # match control: Next word (NW), Previous word (PW), End of line(EOL), Until delimeter (UD)
    removeText=None # remove the given text to generate meta info name
    alsoMatch=None # also match the strings in this list with the given key in same line (any where)
    metaHeader=None # the general meta name header (See xxxDictionary for usage)
    metaName=None # this will be the meta info name that will be passed to backend and written output
    metaNameTag=None # the meta info name body
    metaInfoType=None # Float, string or other that have to be defined in meta info core
    value=None # The final value that will be passed to backend and written to output
    valueSize=None # The size of the value that will be pased to backend and written to output
    sizeMetaName=None
    depends=None # list/dict for altering values in this or other dictionaries
    lookupdict=None # the lookup dict for values in depends (can also be itself)
    subfunction=None # an extra update function before passing value to backend or write output
    activeSections=None # The sections that this meta info key will be active to pass backend
    autoSections=False

    def __init__(self, *args, **kwargs):
        super(MetaInfoMap, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    if k in self:
                        self[k] = v
        if kwargs:
            for k, v in kwargs.items():
                if k in self:
                    self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(MetaInfoMap, self).__setitem__(key, value)
        self.__dict__.update({key: value})

class FileInfoMap(dict):
    """Map cache values to meta info
    """
    activeInfo=False
    infoPurpose=None
    fileName=None
    fileFormat=None
    fileSupplied=False
    fileHolder=None
    nameTranslate=None
    matchStr=None
    matchNames=None
    metaHeader=None
    metaName=None
    metaNameTag=None
    metaInfoType=None
    value=None
    valueSize=None
    sizeMetaName=None
    depends=None
    lookupdict=None
    subfunction=None
    activeSections=None

    def __init__(self, *args, **kwargs):
        super(FileInfoMap, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    if k in self:
                        self[k] = v
        if kwargs:
            for k, v in kwargs.items():
                if k in self:
                    self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(FileInfoMap, self).__setitem__(key, value)
        self.__dict__.update({key: value})

class MapDictionary(dict):
    """
    Modified from the reference source below:
    https://stackoverflow.com/questions/2352181/how-to-use-a-dot-to-access-members-of-dictionary
    Example:
    m = MapDictionary({'Name': 'mdtraj'}, format='.mdcrd', found=True, list=['Value'])
    """
    def __init__(self, *args, **kwargs):
        super(MapDictionary, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    if (isinstance(v, FileInfoMap) or
                        isinstance(v, MetaInfoMap)):
                        if v.replaceTag:
                            k_new = v.replaceTag
                        else:
                            k_new = k
                        if v.nameTranslate:
                            v.metaName = v.nameTranslate(k_new)
                        else:
                            v.metaName = k_new
                    v.matchStr = k
                    metaStr = ''
                    if v.metaHeader:
                        metaStr = metaStr + v.metaHeader + '_'
                    if v.metaNameTag:
                        metaStr = metaStr + v.metaNameTag + '_'
                    metaStr = metaStr + v.metaName
                    self[metaStr] = v
                    if metaStr != k:
                        self.pop(k, None)

        if kwargs:
            for k, v in kwargs.items():
                if (isinstance(v, FileInfoMap) or
                    isinstance(v, MetaInfoMap)):
                    if v.metaTranslate:
                        v.metaName = v.nameTranslate(k)
                    else:
                        v.metaName = k
                v.matchStr = k
                metaStr = ''
                if v.metaHeader:
                    metaStr = metaStr + v.metaHeader + '_'
                if v.metaNameTag:
                    metaStr = metaStr + v.metaNameTag + '_'
                metaStr = metaStr + v.metaName
                self[metaStr] = v
                if metaStr != k:
                    self.pop(k, None)

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(MapDictionary, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(MapDictionary, self).__delitem__(key)
        del self.__dict__[key]

    def get_keys(self):
        return [val.metaName for val in self.__dict__.values()]

def getList_MetaStrInDict(sourceDict):
    """Returns a list that includes all meta name
       strings for the given dictionary.
       Meta name strings are not actual meta names but
       used as the keywords in the parsing.
    """
    return [sourceDict[item].matchStr for item in sourceDict]

def copyMetaDictToDict(sourceObj, sourceDict, sourceType=None, targetObj=None,
                       targetDict=None, targetType=None, copyDict=None):
    """ Copy values of a source dictionary to a target one
        by matching the keys in bith dictionaries.
        Dictionary types can be python or SmartParser

        Inputs:
            sourceObj  (object) : python object that holds sourceDict
            sourceDict (string) : the name of the source dictionary
            sourceType (string) : type can be 'standard' or 'smartparser'
            targetObj  (object) : python object that holds targetDict
                                  if None, sourceObj will be used
            targetDict (string) : the name of the target dictionary
                                  if None, sourceDict will be used
            targetType (string) : type can be 'standard' or 'smartparser'
                                  if targetDict is None, this equals to sourceType
            copyDict (dict)     : includes items for mapping
                                  keys from source to target
                                  Ex. : {"ener0" : "Energy", "step" : "Timestep"}

            If target inputs are None, the copy will work on the same dictionary.

            Outputs: [Success, Status]
            Success :  True if any copy else False
            Status  : -2 if both dictionaries are not accessible
                      -1 if only target dict is not accessible
                       0 can access dict but without any copy
                       1 at least one copy is made from sourceDict to targetDict
                       N number of N copies were made from sourceDict to targetDict
    """
    rtn = [False,-2]
    skip = False
    sDict = None
    tDict = None
    if targetDict is None and copyList is None:
        skip = True
    if skip is False:
        if sourceType is None:
            sourceType = 'smartparser'
        if targetObj is None:
            targetObj = sourceObj
        if targetDict is None:
            targetDict = sourceDict
            targetType = sourceType
        if targetType is None:
            targetType = sourceType
        sDict = getattr(sourceObj, sourceDict)
        tDict = getattr(targetObj, targetDict)
        if sDict is not None:
            rtn[1] += 1
        if tDict is not None:
            rtn[1] += 1
        sp = False
        if 'smartparser' in sourceType:
            sp = True
        tp = False
        if 'smartparser' in targetType:
            tp = True
        for skey, sval in sDict.items():
            sk = sval.matchStr if sp else skey
            if copyDict:
                if sk in copyDict:
                    target = copyDict[sk]
                    for tkey, tval in tDict.items():
                        tk = tval.matchStr if tp else tkey
                        if target == tk:
                            rtn[0] = True
                            rtn[1] += 1
                            if tp:
                                tDict[tkey] = sval.value if sp else sval
                            else:
                                tDict[tkey] = sval.value if sp else sval
            else:
                for tkey, tval in tDict.items():
                    tk = tval.matchStr if tp else tkey
                    if sk == tk:
                        rtn[0] = True
                        rtn[1] += 1
                        if tp:
                            tDict[tkey] = sval.value if sp else sval
                        else:
                            tDict[tkey] = sval.value if sp else sval
                        break
        if rtn[1]>0:
            setattr(targetObj, tDict, targetDict)
    return rtn

def setMetaStrInDict(objectName, sourceDict, mStr, mVal, matchExact=False,
        caseSensitive=False, multipleMatch=True, matchFirst=None, matchLast=None):
    """Find mStr in a given sourceDict and set it to mVal.
       If the value is set in a dictionary item, activeInfo of the
           item is also set to True

       Inputs:
           objectName (String) : Python object that holds this attribute
                                 Ex. : self
           sourceDict (String) : The name string of the SmartParser dictionary
                                 Ex. : 'myDict'  (this will be self.myDict)
           mStr (String) : The string to be matched with the matchStr in dict.
           mVal (Any type that is accepted by metaInfoStorage) : The value
       Output
           A list with [Success, MatchKey, Status]:
               Success:  True : there is at least one match
                         False: there is an error
               MatchKey       : a list with the matched
                                  keys in the dict
               Status :  -1   : there is no dictionary
                                  with the given name
                          0   : there is dict but there is no
                                  item that matchs with mStr
                          1   : there is a match in the dict
                                  with a successful value set
                          2-N : there are multiple matches
                                  with successful value settings
                                  (Only available if multipleMatch is True)
    """
    rtnList = [False, None, -1]
    theDict = getattr(objectName, sourceDict)
    if theDict is not None:
        rtnList[2] = 0
        matchKeys = []
        if matchFirst is not None:
            if int(matchFirst)<1:
                matchFirst = None
            else:
                macthExact = False
        if matchLast is not None:
            if int(matchLast)<1:
                matchLast = None
            else:
                macthExact = False
        for k,v in theDict.items():
            assign = False
            if matchExact is True:
                if caseSensitive:
                    if mStr == v.matchStr:
                        assign = True
                else:
                    if mStr.upper() == v.matchStr.upper():
                        assign = True
            elif matchFirst is not None:
                mFirst = int(matchFirst)
                if mFirst >= len(v.matchStr):
                    mFirst=len(v.matchStr)
                vStr = v.matchStr[0:mFirst]
                if caseSensitive:
                    if mStr in vStr:
                        assign = True
                else:
                    if mStr.upper() in vStr.upper():
                        assign = True
            elif matchLast is not None:
                mLast = int(matchLast)
                if mLast >= len(v.matchStr):
                    mLast=len(v.matchStr)
                vStr = v.matchStr[-mLast:-1]
                if caseSensitive:
                    if mStr in vStr:
                        assign = True
                else:
                    if mStr.upper() in vStr.upper():
                        assign = True
            else:
                if caseSensitive:
                    if mStr in v.matchStr:
                        assign = True
                else:
                    if mStr.upper() in v.matchStr.upper():
                        assign = True
            if assign:
                theDict[k].value = mVal
                theDict[k].activeInfo = True
                matchKeys.append(k)
                if multipleMatch is False:
                    break
        if matchKeys:
            rtnList[0] = True
            rtnList[1] = matchKeys
            rtnList[2] = len(matchKeys)
            setattr(objectName, sourceDict, theDict)
    return rtnList

def isMetaStrInDict(nameStr, sourceDict):
    """Returns a list that includes all meta name
       strings for the given dictionary.
       Meta name strings are not actual meta names but
       used as the keywords in the parsing.
    """
    val = None
    for k,v in sourceDict.items():
        if nameStr in v.matchStr:
            val = k
            break
    return val
    #return [k for k,v in sourceDict.items() if nameStr in v.matchStr][0]

def getDict_MetaStrInDict(sourceDict, nameList=None):
    """Returns a dict that includes all meta name
       strings and corresponding values for the given dictionary.
       Meta name strings are not actual meta names but
       used as the keywords in the parsing.
    """
    newDict = {}
    if nameList is not None:
        if isinstance(nameList, (list, tuple)):
            for key in nameList:
                if key in sourceDict:
                    newDict.update({sourceDict[key].matchStr : key})
        elif isinstance(nameList, dict):
            for key, value in nameList:
                if value in sourceDict:
                    newDict.update({key : value})
        elif isinstance(nameList, str):
            for key, value in sourceDict.items():
                if nameList in sourceDict[key]:
                    if isinstance(sourceDict[key][nameList], (list, tuple)):
                        for item in sourceDict[key][nameList]:
                            newDict.update({item : key})
                    elif isinstance(sourceDict[key][nameList], str):
                        newDict.update({sourceDict[key][nameList] : key})
                    else:
                        newDict.update({sourceDict[key].matchStr : key})
                else:
                    newDict.update({sourceDict[key].matchStr : key})
    else:
        for key, value in sourceDict.items():
            newDict.update({sourceDict[key].matchStr : key})
    return newDict





# Copyright 2015-2018 Massimo Riello, Adam Fekete, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from LAMMPS.BaseClasses import Lattice


# Default
# lattice none 1.0

# class none(Lattice):
#     name = 'none'
#     pass


class Sc(Lattice):
    name = 'sc'
    pass


class Bcc(Lattice):
    name = 'bcc'
    pass


class Fcc(Lattice):
    name = 'fcc'
    pass


class Hcp(Lattice):
    name = 'hcp'
    pass


class Diamond(Lattice):
    name = 'diamond'
    pass


class Sq(Lattice):
    name = 'sq'
    pass


class sq2(Lattice):
    name = 'sq2'
    pass


class hex(Lattice):
    name = 'hex'
    pass


class custom(Lattice):
    name = 'custom'
    pass




# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Parser for creating artificial test, brenchmark, and demonstration data.
'''

import json
import os.path
import numpy as np
import random
from ase.data import chemical_symbols
import numpy
import sys
import time
import os
import signal

from nomad import metainfo
from nomad.datamodel import EntryArchive
from nomad.datamodel.metainfo.common import section_run as Run

from .legacy import Backend
from .parser import Parser, MatchingParser


class ArtificalParser(Parser):
    ''' Base class for artifical parsers based on VASP metainfo. '''
    def __init__(self):
        super().__init__()
        self.backend = None

    def init_backend(self, **kwargs):
        self.backend = Backend(metainfo='vasp', **kwargs)


class EmptyParser(MatchingParser):
    '''
    Implementation that produces an empty code_run
    '''
    name = "parsers/empty"

    def parse(self, mainfile: str, archive: EntryArchive, logger=None) -> None:
        run = archive.m_create(Run)
        run.program_name = self.code_name


class TemplateParser(ArtificalParser):
    '''
    A parser that generates data based on a template given via the
    mainfile. The template is basically some archive json. Only
    '''
    name = 'parsers/template'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from nomad.datamodel.metainfo import m_env as metainfo_env
        self._metainfo_env = metainfo_env
        self.code_name = 'Template'

    def is_mainfile(
            self, filename: str, mime: str, buffer: bytes, decoded_buffer: str,
            compression: str = None) -> bool:
        return filename.endswith('template.json')

    def transform_value(self, name, value):
        ''' allow subclasses to modify values '''
        return value

    def transform_section(self, name, section):
        ''' allow subclasses to modify sections '''
        return section

    def add_section(self, section):
        name = section['_name']
        index = self.backend.openSection(name)

        for key, value in section.items():
            if key.startswith('x_') or key.startswith('_'):
                continue

            if key.startswith('section_'):
                values = value if isinstance(value, list) else [value]
                for value in values:
                    self.add_section(self.transform_section(key, value))
            else:
                value = self.transform_value(key, value)
                if isinstance(value, list):
                    quantity_def = self.backend.env.resolve_definition(key, metainfo.Quantity)
                    if quantity_def.is_scalar:
                        for single_value in value:
                            self.backend.addValue(key, single_value, index)
                    else:
                        self.backend.addArrayValues(key, np.asarray(value), index)
                else:
                    self.backend.addValue(key, value, index)

        self.backend.closeSection(name, index)

    def parse(self, mainfile: str, archive: EntryArchive, logger=None) -> None:
        # tell tests about received logger
        if logger is not None:
            logger.debug('received logger')

        self.init_backend(entry_archive=archive)

        if 'warning' in mainfile:
            self.backend.pwarn('A test warning.')

        template_json = json.load(open(mainfile, 'r'))
        self.add_section(template_json['section_run'][0])
        if 'section_workflow' in template_json:
            self.add_section(template_json['section_workflow'])
        self.backend.finishedParsingSession('ParseSuccess', [])
        logger.debug('a test log entry')


class ChaosParser(ArtificalParser):
    '''
    Parser that emulates typical error situations. Files can contain a json string (or
    object with key `chaos`) with one of the following string values:
    - exit
    - deadlock
    - consume_ram
    - exception
    - segfault
    - random
    '''
    name = 'parsers/chaos'

    def is_mainfile(
            self, filename: str, mime: str, buffer: bytes, decoded_buffer: str,
            compression: str = None) -> bool:
        return filename.endswith('chaos.json')

    def parse(self, mainfile: str, archive: EntryArchive, logger=None) -> None:
        self.init_backend(entry_archive=archive)

        chaos_json = json.load(open(mainfile, 'r'))
        if isinstance(chaos_json, str):
            chaos = chaos_json
        elif isinstance(chaos_json, dict):
            chaos = chaos_json.get('chaos', None)
        else:
            chaos = None

        if chaos == 'random':
            chaos = random.choice(['exit', 'deadlock', 'consume_ram', 'exception', 'segfault'])

        if chaos == 'exit':
            sys.exit(1)
        elif chaos == 'deadlock':
            while True:
                time.sleep(1)
        elif chaos == 'consume_ram':
            data = []
            i = 0
            while True:
                data.append('a' * 10**6)
                i += 1
                logger.info('ate %d mb' % i)
        elif chaos == 'exception':
            raise Exception('Some chaos happened, muhuha...')
        elif chaos == 'segfault':
            os.kill(os.getpid(), signal.SIGSEGV)

        raise Exception('Unknown chaos')


class GenerateRandomParser(TemplateParser):
    name = 'parsers/random'

    basis_set_types = [
        'Numeric AOs', 'Gaussians', '(L)APW+lo', 'Plane waves',
        'Real-space grid', 'Local-orbital minimum-basis']

    electronic_structure_methods = [
        'DFT', 'DFT+U', 'full-CI', 'CIS', 'CISD'
        'CCS', 'CCS(D)', 'CCSD', 'CCSD(T)', 'CCSDT(Q)', 'MP2', 'MP3', 'MP4', 'MP5', 'MP6',
        'G0W0', 'scGW', 'LDA', 'hybrid', 'CASPT2', 'MRCIS', 'MRCISD', 'RAS-CI']

    XC_functional_names = [
        'LDA_X', 'LDA_C', 'GGA_X', 'GGA_C', 'HYB_GGA_XC', 'MGGA_X', 'MGGA_C', 'HYB_MGGA_XC']

    low_numbers = [1, 1, 2, 2, 2, 2, 2, 3, 3, 4]

    def __init__(self):
        super(GenerateRandomParser, self).__init__()
        file_dir = os.path.dirname(os.path.abspath(__file__))
        relative_template_file = "random_template.json"
        template_file = os.path.normpath(os.path.join(file_dir, relative_template_file))
        self.template = json.load(open(template_file, 'r'))
        self.random = None

    def is_mainfile(
            self, filename: str, mime: str, buffer: bytes, decoded_buffer: str,
            compression: str = None) -> bool:
        return os.path.basename(filename).startswith('random_')

    def transform_section(self, name, section):
        ''' allow subclasses to modify sections '''
        if name == 'section_system':
            atoms = []
            atom_positions = []
            # different atoms
            for _ in range(0, random.choice(GenerateRandomParser.low_numbers)):
                # ase data starts with a place holder value X, we don't want that
                atom = random.choice(chemical_symbols[1:])
                # atoms of the same number
                for _ in range(0, random.choice(GenerateRandomParser.low_numbers)):
                    atoms.append(atom)
                    atom_positions.append([.0, .0, .0])

            section['atom_labels'] = atoms
            section['atom_positions'] = atom_positions
            return section
        else:
            return section

    def transform_value(self, name, value):
        if name == 'program_basis_set_type':
            return random.choice(GenerateRandomParser.basis_set_types)
        elif name == 'electronic_structure_method':
            return random.choice(GenerateRandomParser.electronic_structure_methods)
        elif name == 'XC_functional_name':
            return random.choice(GenerateRandomParser.XC_functional_names)
        elif name == 'atom_positions':
            return [numpy.random.uniform(0e-10, 1e-9, 3) for _ in value]
        else:
            return value

    def parse(self, mainfile: str, archive: EntryArchive, logger=None) -> None:
        # tell tests about received logger
        if logger is not None:
            logger.debug('received logger')

        self.init_backend(entry_archive=archive)
        seed = int(os.path.basename(mainfile).split('_')[1])
        random.seed(seed)
        numpy.random.seed(seed)
        section = self.template['section_run'][0]
        self.add_section(section)
        self.backend.finishedParsingSession('ParseSuccess', [])



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine, exc
from sqlalchemy.orm import sessionmaker
import logging, sys

logger = logging.getLogger(__name__)
# handler = logging.StreamHandler(stream=sys.stdout)
# logger.setLevel(logging.INFO)
# logger.addHandler(handler)

Base = declarative_base()
useNested = False
useJson = False

if useJson:
    from sqlalchemy.dialects.postgresql import JSON, JSONB

def compareSorted(oldValues, newValues):
    toAdd, toRemove = [], []
    iNew, iOld = 0, 0
    while iNew < len(newValues) and iOld < len(oldValues):
        if newValues[iNew] < oldValues[iOld]:
            toAdd.append(newValues[iNew])
            iNew += 1
        elif newValues[iNew] > oldValues[iOld]:
            toRemove.append(oldValues[iOld])
            iOld += 1
        else:
            iNew += 1
            iOld += 1
    toAdd += newValues[iNew:]
    toRemove += oldValues[iOld:]
    return (toRemove, toAdd)

def createEngine(engineStr='sqlite:///:memory:', echo = True):
    return create_engine(engineStr, echo = echo)

def createDB(engine):
    Base.metadata.create_all(engine)

def createSession(engine):
    return sessionmaker(bind=engine)()

def get_or_create(cls, session, defaults=None, **kwds):
    result = session.query(cls).filter_by(**kwds).first()
    if result:
        return result, True
    newVals=defaults
    if defaults is None:
        newVals={}
    newVals.update(kwds)
    result = cls(**newVals)
    if useNested:
        nestedSession = session.begin_nested()
        nestedSession.add(result)
        try:
            nestedSession.commit()
        except exc.IntegrityError:
            nestedSession.rollback()
            result = session.query(cls).filter_by(**kwds).one()
    else:
        session.add(result)
        session.flush()
    return result, True



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nomad.metainfo import MSection, Quantity


class Section(MSection):
    foo = Quantity(name='foo', type=int)


bar_section = Section()
bar_section.foo = 1


class Regular:
    def __init__(self):
        self.foo = 1


obj = Regular()


class Property:
    def __init__(self):
        self.foo = 1

    @property
    def foo(self,):
        return self._foo

    @foo.setter
    def foo(self, value):
        self._foo = value


prop = Property()

dct = dict(foo=1)


def mi(iterations=10):
    bar_section.foo = 1
    for _ in range(0, iterations):
        bar_section.foo += bar_section.foo


def python_obj(iterations=10):
    obj.foo = 1
    for _ in range(0, iterations):
        obj.foo += obj.foo


def python_dct(iterations=10):
    dct['foo'] = 1
    for _ in range(0, iterations):
        dct['foo'] += dct['foo']


def python_property(iterations=10):
    prop.foo = 1
    for _ in range(0, iterations):
        prop.foo += prop.foo


def test_mi(benchmark):
    benchmark(mi)


def test_python_obj(benchmark):
    benchmark(python_obj)


def test_python_dct(benchmark):
    benchmark(python_dct)


def test_python_prop(benchmark):
    benchmark(python_property)
